@@ -79,6 +79,7 @@ static DB *db = 0;
 static DBT dbkey = { 0 };
 static DBT dbdata = { 0 };
 static struct triplet_data triplet_data;
+static int deadlock_detect = 0;
 
 static unsigned long greylist_delay = DELAY_MAIL_SECS;
 static unsigned long bloc_max_idle = AUTO_RECORD_LIFE_SECS;
@@ -161,6 +162,11 @@ prepare_env()
 	log_db_error("db_env_create", rc);
     else {
 	dbenv->set_errcall(dbenv, db_errcall_fcn);
+	rc = dbenv->set_lk_detect(dbenv, DB_LOCK_YOUNGEST);
+	if (rc)
+	    log_db_error("dbenv->set_lk_detect DB_LOCK_YOUNGEST, expired triplets will not be deleted", rc);
+	else
+	    deadlock_detect = 1;
 	rc = dbenv->open(dbenv, db_home,
 			 DB_INIT_LOCK | DB_INIT_MPOOL | DB_CREATE, 0);
 	if (rc) {
@@ -238,27 +244,19 @@ cleanup()
 	free(dbkey.data);
     if (db) {
 	rc = db->close(db, 0);
-	db = 0;
 	if (rc)
 	    log_db_error("DB close", rc);
     }
+    rc = db_create(&db, dbenv, 0);
+    if (!rc)
+	if (! db->remove(db, "stats.db", 0, 0))
+	    syslog(LOG_NOTICE, "Unused database stats.db removed");
+    db = 0;
     if (dbenv) {
 	rc = dbenv->close(dbenv, 0);
 	dbenv = 0;
 	if (rc)
 	    log_db_error("DB_ENV close", rc);
-	/* Clean up the environment so that upgrading is easier. */
-	rc = db_env_create(&dbenv, 0);
-	if (rc)
-	    log_db_error("db_env_create failed during cleanup", rc);
-	else {
-	    muffle_error++;
-	    rc = dbenv->remove(dbenv, db_home, 0);
-	    if (rc)
-		syslog(LOG_DEBUG, "DEBUG: BDB-%d: db_env->remove returned: %s",
-		       rc, db_strerror(rc));
-	    muffle_error--;
-	}
     }
     policy_cleanup();
 }
@@ -277,7 +275,10 @@ run_expiry()
     int rc;
     time_t now;
     unsigned int count = 0;
-    if (db == 0)
+    /* Cursor operations can hold several locks and therefore deadlock
+       so don't run expiry if deadlock detection does not work
+       http://www.oracle.com/technology/documentation/berkeley-db/db/ref/lock/notxn.html */
+    if (db == 0 || deadlock_detect == 0)
 	return;
     if (time(&now) == (time_t)-1) {
 	syslog(LOG_ERR, "time failed during run_expiry");
@@ -311,7 +312,10 @@ run_expiry()
 		    count++;
 	    }
 	}
-	if (rc != DB_NOTFOUND)
+	if (rc == DB_LOCK_DEADLOCK)
+	    syslog(LOG_DEBUG, "skipping concurrent expiry avoids "
+		   "deadlocks and unnecessary work");
+	else if (rc != DB_NOTFOUND)
 	    log_db_error("dbcp->c_get failed", rc);
 	if (rc = dbcp->c_close(dbcp))
 	    log_db_error("dbcp->c_close failed", rc);
@@ -501,13 +505,14 @@ get_grey_data()
     }
 }
 
-static void
+static int
 put_grey_data()
 {
     int rc;
     rc = db->put(db, NULL, &dbkey, &dbdata, 0);
     if (rc)
 	log_db_error("put", rc);
+    return rc;
 }
 
 static void
@@ -550,7 +555,7 @@ printf_action(const char *fmt, unsigned long delay)
     }
 }
 
-static void
+static int
 process_smtp_rcpt()
 {
     double delay;
@@ -561,7 +566,7 @@ process_smtp_rcpt()
 	}
 	else
 	    puts(STR_ACTION "WARN " PACKAGE_STRING " is not working properly");
-	return;
+	return 1;
     }
     get_grey_data();
     delay = difftime(triplet_data.access_time, triplet_data.create_time);
@@ -579,7 +584,7 @@ process_smtp_rcpt()
 	printf_action(greylisted_action_fmt, delay);
 	putchar('\n');
     }
-    put_grey_data();
+    return put_grey_data();
 }
 
 static void
@@ -824,7 +829,12 @@ main(int argc, const char **argv)
 	    && (from = find_attribute("sender"))
 	    && (to = find_attribute("recipient"))) {
 	    build_triplet_key(ip, from, to);
-	    process_smtp_rcpt();
+	    if (process_smtp_rcpt()) {
+	      syslog(LOG_ERR, "error: cleaning up");
+	      cleanup();
+	      syslog(LOG_ERR, "error: exiting after cleanup");
+	      return 1;
+	    }
 	}
 	else {
 	    puts(STR_ACTION "DUNNO");
